from __future__ import print_function
from .logsetup import UserLogger
import requests
import ssl
import sys
from infi.credentials_store import CLICredentialsStore, Credentials
import pickle
import traceback
from docopt import docopt
import os
import errno
import sys
import re
from shutil import copy2, rmtree
import grp

requests.packages.urllib3.disable_warnings()

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context


log = UserLogger(__name__)


def handle(signum, frame):
    sys.exit(1)


def get_function_name():
    return traceback.extract_stack(None, 2)[0][2]


def load_cookie(infinibox, username, hostname):
    try:
        with open('/tmp/iguard_tools_cookie_{}_{}'.format(username, hostname), 'rb') as f:
            return pickle.load(f)
    except Exception:
        log.debug("Cookie not found.")
    return None


def save_cookie(infinibox, username, hostname):
    cookie = None
    try:
        with open('/tmp/iguard_tools_cookie_{}_{}'.format(username, hostname), 'wb') as f:
            pickle.dump(infinibox.api._session.cookies.copy(), f)
    except Exception as e:
        log.debug("Failed to save cookie")
        log.debug(repr(e))


def my_login(infinibox, username, hostname):
    cookie = load_cookie(infinibox, username, hostname)
    if not cookie:
        log.debug("Cookie not found. Logging in.")
        infinibox.login()
        cookie = save_cookie(infinibox, username, hostname)
    else:
        log.debug("Loaded saved cookie")
        infinibox.api._session.cookies = cookie
        try:
            infinibox.get_version()
            # It didn't use the cookie, but looged in
            if infinibox.is_logged_in():
                log.debug("Logged in, cookie expired. Saving.")
                save_cookie(infinibox, username, hostname)
        except Exception:
            log.debug("Cookie not accepted")
            infinibox.api._session.cookies.clear()
            infinibox.login()
            save_cookie(infinibox, username, hostname)


class MyCLICrdentialsStore(CLICredentialsStore):
    def _get_file_folder(self):
        return ".infi.credentials_store"

    def authenticate(self, key, credentials):
        return True

    def ask_credentials_prompt(self, key):
        print('Connecting to system ' + str(key))


def delete_credentials(store):
    print("Deleting all credentials")
    store.reset_credentials()


def save_credentials(store, hostname, username, password):
    credentials = Credentials(username, password)
    store.set_credentials(hostname, credentials)
    print("Stored credentials for host {}".format(hostname))


def validate_config(arguments, store):
    config = {}
    if arguments['--help']:
        docopt(__doc__, '-h')

    for key, value in arguments.items():
        config[key[2:]] = value
    if not config['delete']:
        if not config['username'] or not config['password']:
            credentials = store.get_credentials(config['hostname'])
            config['username'] = credentials.get_username()
            config['password'] = credentials.get_password()
    if not config['delete'] and not config['save']:
        try:
            dde_id = int(config['dde'])
            if dde_id not in [1, 2]:
                raise Exception
            config['dde'] = dde_id
        except Exception as e:
            raise Exception('DDE ID should be 1 or 2. Exiting.')
    return config


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_err(*args, **kwargs):
    print(f'{bcolors.FAIL}{" ".join([a for a in args])}{bcolors.ENDC}', file=sys.stderr, **kwargs)


def is_task_finished(infinibox, dde_id, task_id):
    task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
    get = infinibox.api.get(task_url)
    status = get.get_result().get('status')
    return status != 'RUNNING'


def post_install():
    print(sys.argv)
    ORACLE_HOME = os.environ.get('ORACLE_HOME')
    if not ORACLE_HOME:

        print('Error: ORACLE_HOME must be defined')
        print()
        sys.stdout.flush()
        sys.exit(1)
    if not os.path.isdir(ORACLE_HOME):
        print('Error: ORACLE_HOME directory {} does not exist!'.format(ORACLE_HOME))
        print('Verify your Oracle installation is correct.')
        sys.exit(1)
    if not os.path.isdir(ORACLE_HOME + '/lib'):
        print('Error: ORACLE_HOME lib directory {}/lib does not exist!'.format(ORACLE_HOME))
        print('Verify your Oracle installation is correct.')
        sys.exit(1)

    os_name = os.uname()[0].lower()
    if os_name == 'aix':
        with open(ORACLE_HOME + '/rdbms/lib/config.s', "r") as f:
            for line in f:
                if line.find(".string") != -1:
                    oracle_group = re.sub(r'["\'\t]', '', line.split()[1])
    else:
        with open(ORACLE_HOME + '/rdbms/lib/config.c', "r") as f:
            for line in f:
                if line.find("define SS_DBA_GR") != -1:
                    oracle_group = re.sub(r'["\'\t]', '', line.split()[2])

        if not oracle_group:
            print("Error: Could not determine Oracle OSDBA group.")
            print("Verify your Oracle installation is correct.")
            sys.exit(1)

    files = {
        '/usr/INFINIDAT/libINFINIDATobk.so': 'libINFINIDATobk.so',
        '/usr/INFINIDAT/key.pem': 'key.pem',
        '/usr/INFINIDAT/INFINIDATPlugin.conf': 'INFINIDATPlugin.conf',
        '/usr/INFINIDAT/EULA.txt': 'EULA.txt',
        '/usr/INFINIDAT/cert.pem': 'cert.pem',
        '/usr/INFINIDAT/cacert.pem': 'cacert.pem',
        '/usr/bin/rman_password_encrypter': 'rman_password_encrypter',
        '/usr/INFINIDAT/log/': '',
        '/usr/INFINIDAT/creds/': '',
    }
    cwd = os.path.dirname(__file__) + '/files/'
    for target, source in files.items():
        base = os.path.dirname(target)
        os.makedirs(base, exist_ok=True)
        if source:
            copy2('{}{}'.format(cwd, source), target)

    symlink_path = '{}/lib/libINFINIDATobk.so'.format(ORACLE_HOME)
    try:
        os.symlink('/usr/INFINIDAT/libINFINIDATobk.so', symlink_path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(symlink_path)
            os.symlink('/usr/INFINIDAT/libINFINIDATobk.so', symlink_path)
        else:
            print("Could not create symlink {}. Exiting.".format(symlink_path))
            sys.exit(1)

    try:
        gid = grp.getgrnam(oracle_group).gr_gid
    except KeyError:
        print("Group {} not found. Exiting.".format(oracle_group))
        sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)

    os.chown('/usr/INFINIDAT', -1, gid)
    for root, dirs, files in os.walk('/usr/INFINIDAT'):
        for name in dirs:
            os.chown(os.path.join(root, name), -1, gid)
        for name in files:
            os.chown(os.path.join(root, name), -1, gid)

    msg = '''
Installation successfully completed.

A link to the INFINIDAT RMAN lib was created in: {ORACLE_HOME}/lib
{ORACLE_HOME}/lib/libINFINIDATobk.so -> /usr/INFINIDAT/libINFINIDATobk.so

If you have other Oracle Database installations, you will need to create similar links
in each unique ORACLE_HOME/lib location.


See the INFINIDAT InfiniGuard RMAN Plug-in README document for further information on configuring RMAN support.


See the Client Plugin End User License Agreement in /usr/INFINIDAT/EULA.txt.
'''.format(ORACLE_HOME=ORACLE_HOME)
    print(msg)


def pre_uninstall():
    try:
        rmtree('/usr/INFINIDAT')
        os.remove('/usr/bin/rman_password_encrypter')
    except:
        pass
