"""

Usage:
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> 
                [--age=<age>][--share=<share>][--long][--time|--size][--reverse][--human-readable]
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> --list
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> --replication
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> --netboost
   iguard_ls --hostname=<hostname> --username=<username> --password=<password> --save
   iguard_ls --delete
   iguard_ls --help
   iguard_ls --version


Options:
   -H hostname --hostname=<hostname>   DNS name/IP  of InfiniGuard
   -u username --username=<username>   Administrator's username
   -p password --password=<password>   Administrator's password
   -d dde_id, --dde=<dde_id>           DDE ID (1/2)
   -s share, --share=<share>           Display only share name <share>
   -a age, --age=<age>                 Only display files older than <age> days
   -D, --delete                        Delete all stored credentials
   -S, --save                          Save credentials for hostname
   -L, --list                          List all share names
   -l, --long                          Use long display format
   -t, --time                          Sort by time
   --size                              Sort by size
   -r, --reverse                       Sort in reverse order
   -R, --replication                   Show file/dir replication status
   -N, --netboost                      Show NetBoost real-time statistics
   -h, --human-readable                Display size in human readable format
   --help                              Show this screen
   -V, --version                       Show iguard_ls version
"""
from __future__ import print_function
from .logsetup import UserLogger, setup_logging
from docopt import docopt
from .__version__ import __version__ as iguard_ls_version
from infinisdk import InfiniBox
import requests
import sys
from time import asctime, time, localtime
from datetime import datetime
import stat
from infinisdk.core.exceptions import APICommandFailed
from time import sleep
import signal
import re
from infiniguard_tools.common import (
    handle,
    MyCLICrdentialsStore,
    validate_config,
    my_login,
    get_function_name,
    is_task_finished,
    print_err,
    delete_credentials,
    save_credentials
)

requests.packages.urllib3.disable_warnings()

log = UserLogger(__name__)


def _print_files(share_data, app_share, is_long=False, human_readable=False):
    def human_readable_size(size, decimal_places=1):
        for unit in ['', 'K', 'M', 'G', 'T', 'P']:
            if size < 1024.0:
                break
            size /= 1024.0
        if not unit:
            return "{}".format(size)
        else:
            return "{size:.{decimal_places}f}{unit}".format(size=size, decimal_places=decimal_places, unit=unit)

    for file in share_data:
        path = file.get('filename')
        if is_long:
            size = file.get('size')
            mtime = asctime(localtime(file.get('mtime')))
            mode = stat.filemode(file.get('mode'))
            uid = file.get('uid')
            gid = file.get('gid')
            if human_readable:
                size_format = '{:>7}'.format(human_readable_size(size))
            else:
                size_format = '{:<12}'.format(size)
            print('{} {:<4} {:<4} {} {} {}/{}'.format(mode,
                                                      uid, gid, mtime, size_format, app_share, path))
        else:
            print('{}/{}'.format(app_share, path))


def get_shares(config):
    infinibox = config.get('infinibox')
    dde_id = config.get('dde')
    get = infinibox.api.get(
        "/api/dde/{}/api/nas/app_specific/shares/".format(dde_id))
    res = get.get_result()
    return [a.get('name') for a in res]


def print_shares(config):
    for share in get_shares(config):
        print(share)


def match_netboost_stats(metric):
    if re.match(r'ost_aggregate_.*', metric):
        return True
    if re.match(r'ost_\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}.*', metric):
        return True
    return False


def get_netboost_stats(config):
    infinibox = config.get('infinibox')
    dde_id = config.get('dde')
    get = infinibox.api.get(
        "/api/dde/{}/stats/performance/stats".format(dde_id))
    res = get.get_result()
    fields = res.get('fields')
    data = res.get('data')[0]
    ret = {k: v for k, v in zip(fields, data) if match_netboost_stats(k)}
    return ret


def print_netboost_stats(config):
    def human_readable_size(size, decimal_places=0):
        for unit in ['', 'KB', 'MB', 'GB', 'TB', 'PB']:
            if size < 1000.0:
                break
            size /= 1000.0
        if not unit:
            return "{}".format(size)
        else:
            return "{size:.{decimal_places}f} {unit}/sec".format(size=size, decimal_places=decimal_places, unit=unit)

    netboost_stats = get_netboost_stats(config)
    ret = dict()
    for k, v in netboost_stats.items():
        entity = k.split('_')[1]
        if entity not in ret:
            ret[entity] = dict()
        metric = "_".join(k.split('_')[2:])
        ret[entity][metric] = v
    if ret:
        for client, data in ret.items():
            if client == 'aggregate':
                client = 'Total'
            ethernet_in = data.get('ethernet_in')
            inline = data.get('inline')
            bw_reduction = data.get('bandwidth_reduction')
            dedup_ratio = inline // ethernet_in
            print("{}:".format(client))
            print("{:>25}  {:<11}".format(
                "Ethernet In:", human_readable_size(ethernet_in)))
            print("{:>25}  {:<11}".format(
                "Inline Speed:", human_readable_size(inline)))
            print("{:>25}  1:{:<4}".format("Dedupe Ratio:", dedup_ratio))
            print("{:>25}  {:<5}%".format(
                "Bandwidth Reduction:", bw_reduction))

            if client == 'Total':
                print("---------------------------------------")
            print()
    else:
        print("No backups are currently running.")


def get_replication(config):
    infinibox = config.get('infinibox')
    dde_id = config.get('dde')
    get = infinibox.api.get(
        "/api/dde/{}/stats/replication".format(dde_id))
    res = get.get_result()
    return [a for a in res if a.get('replication_type') == 'FILEDIR']


def print_replication(config):
    replication_list = get_replication(config)
    if not replication_list:
        print("No filesystems with configured file/dir replication found!")
    for replication in replication_list:
        name = replication.get('name')
        utc_completion_time = replication.get('estimated_completion_time')
        average_transfer_rate = replication.get('average_transfer_rate')
        remaining_data_to_transfer = replication.get(
            'remaining_data_to_transfer')
        replication_state = replication.get('replication_state')
        estimated_duration = replication.get('estimated_duration')
        last_replicated = replication.get('last_transferred', "Unsupported")

        print("{}:".format(name))
        print("{:>35}\t{}".format("State:", replication_state))
        if last_replicated != "Unsupported":
            replicated_time = "Never" if last_replicated == "0" else asctime(
                localtime(int(last_replicated) // 1000))
            print("{:>35}\t{}".format(
                "Last replicated file timestamp:", replicated_time))

        if remaining_data_to_transfer == "0" and replication_state == "Enabled":
            print("{:>35}\tFilesystem is syncronized.".format("Remaining:"))
        else:
            # Format is Wed Mar 31 22:07:52 2021

            datetime_completion_time = datetime.strptime(
                utc_completion_time, '%a %b %d %H:%M:%S %Y') - datetime(1970, 1, 1)
            completion_time = asctime(
                localtime(datetime_completion_time.total_seconds()))
            print("{:>35}\t{}".format(
                "Remaining:", remaining_data_to_transfer))
            print("{:>35}\t{}".format("Completion:", completion_time))
            print("{:>35}\t{}".format("Speed:", average_transfer_rate))
            print("{:>35}\t{}".format("Duration:", estimated_duration))


def get_files(infinibox, dde_id, task_id, last_id):
    result_url = "/api/dde/{}/api/asynctasks/{}/file_results?sort=id&page_size=1000&page=1&id=gt:{}".format(
        dde_id, task_id, last_id)
    get = infinibox.api.get(result_url)
    res = get.get_result()
    return res


def print_files(config):
    log.info('Running {}'.format(get_function_name()))

    dde_id = config['dde']
    infinibox = config['infinibox']
    EXCLUDES = ['infinibox', 'version']
    data = {k: v for k, v in config.items(
    ) if k not in EXCLUDES and v is not None}
    log.debug(data)
    is_long = config.get('long')
    is_human_readable = config.get('human-readable')
    is_time = config.get('time')
    is_size = config.get('size')
    is_reverse = True if config.get('reverse') else False
    is_sorted = is_time or is_size
    share = config.get('share', None)
    shares = get_shares(config)
    if share:
        if share in shares:
            shares = [share]
        else:
            print('Share {} not found on DDE. Exiting.'.format(share))
            sys.exit(1)
    age = config.get('age', 0)
    if not age:
        age_query = ''
    else:
        now = int(time())
        seconds = int(age) * 3600 * 24
        age_query = '?mtime=lte:{}'.format(now - seconds)

    log.info("Shares: {}".format(shares))
    rc = 0
    for share in shares:
        post = infinibox.api.post(
            "/api/dde/{}/api/nas/app_specific/shares/{}/list_files{}".format(dde_id, share, age_query))

        task_id = post.get_result().get('task_id')
        task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
        log.debug(task_url)
        result = []
        try:
            last_id = 0
            while True:
                finished = is_task_finished(infinibox, dde_id, task_id)
                out = get_files(infinibox, dde_id, task_id, last_id)
                if out:
                    last_id = out[-1].get('id')
                    if not is_sorted:
                        _print_files(out, share, is_long, is_human_readable)
                    else:
                        result += out
                if finished and not out:
                    break
                else:
                    sleep(0.1)

            if is_sorted:
                if is_size:
                    sortfield = 'size'
                else:
                    sortfield = 'mtime'

                out = sorted(
                    result, key=lambda k: k[sortfield], reverse=not is_reverse)
                _print_files(out, share, is_long, is_human_readable)

        finally:
            delete = infinibox.api.delete(task_url)
            log.debug('Delete status code: {}'.format(delete.status_code))

    sys.exit(rc)


def infiniguard_ls():
    signal.signal(signal.SIGINT, handle)
    try:
        log_handler = setup_logging()
        with log_handler.applicationbound():
            arguments = docopt(__doc__, version=iguard_ls_version)
            store = MyCLICrdentialsStore("infiniguard")
            config = validate_config(arguments, store)
            hostname = config['hostname']
            username = config['username']
            password = config['password']
            dde = config['dde']
            if not config['save'] and not config['delete']:
                infinibox = InfiniBox(hostname, use_ssl=True,
                                      auth=(username, password))
                my_login(infinibox, username, hostname)
                config['infinibox'] = infinibox
                get = infinibox.api.get(
                    '/api/dde/{}/api/asynctasks/'.format(dde))
                if not get.response.content:
                    print_err('This system does not support iguard_ls.')
                    sys.exit(1)

            if config['delete']:
                delete_credentials(store)
            elif config['save']:
                save_credentials(store, hostname, username, password)
            elif config['list']:
                print_shares(config)
            elif config['replication']:
                print_replication(config)
            elif config['netboost']:
                print_netboost_stats(config)
            else:
                print_files(config)

    except APICommandFailed as e:
        if e.status_code == 404:
            print_err('This system does not support iguard_ls')
        elif e.status_code == 503:
            print_err('System Unreachable')
        else:
            print_err(f'ERROR: {e.message}')
        sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)
